      program breakeven
      implicit real*8(a-h,o-z)
      parameter (jsmax=40000,kbmax=400,icfmax=400000)

c     This program computes the break even point for the
c     fast Laplace transform.
c     Given accuracy toler and bounds for the targets 
c     and sources, it finds the values nsorc and mtarg for which
c     the fast laplace transform produces the answer faster than 
c     the direct calculation within the specified error bound.
c     Compile with the line:
c     f77 -o br breakeven.f lapsub.o second.o

c     To do this, it starts with nsorc, mtarg, nterms and q given,
c     and reduces q by qstep until q=1+qstep or the specified error bound
c     and time is achieved. If q=1+qstep when this process terminates,
c     the number of terms is increased by ntstep and the process is repeated.
c     If nterms reaches ntmax without the time and error bound being achieved,
c     nsorc and mtarg are incremented by nstep and the process repeated.
c     Finally it writes the breakeven values of nsorc,mtarg,q,nterms,
c     and the error achieved.

      smin=0.1
      smax=10.0
      tmin=0.1
      tmax=10.0

      write(6,*) ' enter smin,smax,tmin,tmax'
      write(13,*) ' enter smin,smax,tmin,tmax'
      read(5,*) smin,smax,tmin,tmax
      write(6,*) smin,smax,tmin,tmax
      write(13,*) smin,smax,tmin,tmax

      nsorc=100
      mtarg=100 
      nterms=7
      q=1.250d0
      toler=1.d-9
      write(6,*) ' enter nsorc,mtarg,nterms,q,toler,to start'
      write(13,*) ' enter nsorc,mtarg,nterms,q,toler,to start'
      read(5,*) nsorc,mtarg,nterms,q,toler
      if(nsorc.le.0)goto 1000
      write(6,*) nsorc,mtarg,nterms,q,toler
      write(13,*) nsorc,mtarg,nterms,q,toler

      qstep=.020d0
      ntstep=1
      nstep=10
      ntmax=9
      write(6,*) ' enter steps for q,nterms,nsorc and max of nterms'
      write(13,*) 'enter steps for q,nterms,nsorc and max of nterms'
      read(5,*)  qstep,ntstep,nstep,ntmax
      write(6,*)  qstep,ntstep,nstep,ntmax
      write(13,*)  qstep,ntstep,nstep,ntmax

      write(6,887)
      write(13,887)

887   format(7x,'nsorc',1x,'nterms',2x,'q',7x,'tinit',6x,'tfast',
     >       6x,'tslow',5x,'error/Q',7x,'errbd')

      qstart=q
      ntstart=nterms
      nsorc1=nsorc
      mtarg1=mtarg

1     continue
2     continue
3     continue
4     continue

c     Compute error bound first : we only transform if error
c     is GUARANTEED to be small enough
      z=q-1.0d0/q
      th=z/(2.0d0-z)
      errbd=2.0d0*th**(nterms+1)
ccc      errbd=max(errbd,toler)


      tslow=1.0d0
      tfast=1.20d0*tslow
      if(errbd .le. toler)then
      call laplace(smin,smax,tmin,tmax,
     >             nsorc,mtarg,nterms,q,toler,
     >             tinit,tfast,teval,tslow,error,errbd,sumstr,
     >             nsbox,mtbox,kcut,nactb)

c     Check that array bashing doesn't happen
      ntpsq=(nterms+1)**2
      if(nsbox*ntpsq .gt. icfmax .or.
     >   mtbox*ntpsq .gt. icfmax .or.
     >   nsbox .gt. kbmax .or.
     >   mtbox .gt. kbmax)then
      write(6,*) ' error in laplace transform : arrays too small '
      stop
      end if

ccc      if(error.le.toler.and.tfast.le.1.10d0*tslow)goto 1000
ccc      if(errbd.le.toler.and.tfast.le.1.10d0*tslow)goto 1000
      if(tfast.le.1.10d0*tslow)goto 1000

      write(6,888) nsorc,nterms,q,tinit,tfast,tslow,
     >              error/sumstr,errbd
      write(13,888) nsorc,nterms,q,tinit,tfast,tslow,
     >              error/sumstr,errbd
888   format(4x,i8,1x,i4,1x,f6.2,
     >       1x,f10.4,1x,f10.4,1x,f10.4,
     >       1x,e12.4,1x,e12.4)

      end if

      if(q.ge.1.00020d0+qstep)then
      q=q-qstep
      goto 4
      end if
      q=qstart

      if(nterms.le.ntmax)then
      nterms=nterms+ntstep
      goto 3
      end if
      nterms=ntstart

      if(nsorc.le.jsmax)then
      nsorc=nsorc+nstep
      mtarg=mtarg+nstep
      write(6,*) '  '
      write(13,*) '  '
      write(6,887)
      write(13,887)
      goto 2
      end if

1000  continue
      write(6,*) 
      write(6,*) ' breakeven point is at '
      write(6,*) ' nsorc,mtarg,q,nterms '
      write(6,*) nsorc,mtarg,q,nterms
      write(6,*) ' error,error/Q,errbd'
      write(6,*) error,error/sumstr,errbd
      write(6,*) ' tinit,tfast,tslow '
      write(6,*) tinit,tfast,tslow

      write(13,*) 
      write(13,*) ' breakeven point is at '
      write(13,*) ' nsorc,mtarg,q,nterms '
      write(13,*) nsorc,mtarg,q,nterms
      write(13,*) ' error,error/Q,errbd'
      write(13,*) error,error/sumstr,errbd
      write(13,*) ' tinit,tfast,tslow '
      write(13,*) tinit,tfast,tslow

      write(6,*)
      write(13,*)
      write(6,*) ' enter number of cases to run now '
      write(13,*) ' enter number of cases to run now '
      read(5,*) ncase 
      write(6,*) ' enter factor to increase at each case '
      write(13,*) ' enter factor to increase at each case '
      read(5,*) nfact
      nsorc=nsorc/nfact
      mtarg=mtarg/nfact

      write(6,887)
      write(13,887)

      do 100 nc=1,ncase
      nsorc=nsorc*nfact
      mtarg=mtarg*nfact

      call laplace(smin,smax,tmin,tmax,
     >             nsorc,mtarg,nterms,q,toler,
     >             tinit,tfast,teval,tslow,error,errbd,sumstr,
     >             nsbox,mtbox,kcut,nactb)

      write(6,889) nsorc,nterms,q,tinit,tfast,tslow,
     >              error/sumstr,errbd
      write(13,889) nsorc,nterms,q,tinit,tfast,tslow,
     >              error/sumstr,errbd
889   format(4x,i8,1x,i4,1x,f6.2,
     >       1x,f10.4,1x,f10.4,1x,f10.4,
     >       1x,e12.4,1x,e12.4)

100   continue
      

      end

      subroutine laplace(smin,smax,tmin,tmax,
     >                   nsorc,mtarg,nterms,q,toler,
     >                   tinit,tfast,teval,tslow,error,errbd,sumstr,
     >                   nsbox,mtbox,kcut,nactb)
c     Computes fast Laplace transform using symmetric series.

      implicit real*8(a-h,o-z)
      parameter (jsmax=40000,kbmax=400,icfmax=400000)
      real*8 source(jsmax),strength(jsmax),
     >       target(jsmax),transform(jsmax),
     >       coeffs(icfmax),coefft(icfmax),exact(100)
      integer iont(-kbmax:kbmax),ions(-kbmax:kbmax),itbox(jsmax)

c     Generate random sources and targets in specified range
      
ccc      write(6,*) ' Generating points '
      iseed=3474133+nterms+1000*q
      iseed=iseed/2+1
      call mkpoints(tmin,tmax,mtarg,target,strength,iseed)
      iseed=376777+nterms+1000*q
      iseed=iseed/2+1
      call mkpoints(smin,smax,nsorc,source,strength,iseed)

c     Begin timing fast algorithm
c     To ensure accurate timing, do the calculation several times
c     and average if nsorc is small.
      ntimes=1
      if(nsorc.le.20)ntimes=8
      
      t1=second()
      do 954 itime=1,ntimes

c     Form coefficients
      
ccc      write(6,*) ' Forming coefficients'
      call formcf(smin,smax,nsorc,nterms,q,kbsmin,kbsmax,
     >            source,strength,coeffs,ions)

      nsbox=kbsmax-kbsmin
ccc      write(6,*) ' number of source boxes is ',nsbox
ccc      write(13,*) ' number of source boxes is ',nsbox

c      do 10 n=1,100
c      write(6,*) coeffs(n)
c10    continue

c     Assign targets to boxes
      
ccc      write(6,*) ' Assigning targets to boxes'
      call assign(q,kctmin,kctmax,tmin,tmax,mtarg,
     >            target,itbox,iont)

      mtbox=kctmax-kctmin
ccc      write(6,*) ' number of target boxes is ',mtbox
ccc      write(13,*) ' number of target boxes is ',mtbox

ccc      nactb=0
ccc      do 123 kbox=kctmin,kctmax
ccc      nactb=nactb+iont(kbox)
ccc123   continue
ccc      write(6,*) ' number of active target boxes is ',nactb
ccc      write(13,*) ' number of active target boxes is ',nactb

c     Transform coeffs to coefft and accumulate

ccc      write(6,*) ' Transforming coefficients '
      call transt(kbsmin,kbsmax,kctmin,kctmax,nterms,q,toler,
     >               coeffs,coefft,ions,iont,kcut)

954   continue


      temp=second()
      tinit=temp-t1
c     Evaluate expansion at each target

      do 955 itime=1,ntimes
ccc      write(6,*) ' Evaluating expansions'
      call eval(mtarg,nterms,kctmin,kctmax,q,
     >                target,coefft,iont,transform,itbox)
955   continue

c     Finish timing fast algorithm

      t2=second()
      tfast=t2-t1
      teval=t2-temp

      imax=min(100,mtarg)

c     Evaluate directly and compare to fast transform

      t3=second()
ccc      write(6,*) ' Direct evaluation '
      do 956 itime=1,ntimes
      call direct(nsorc,imax,source,strength,target,exact)
956   continue
      t4=second()
      tslow=(t4-t3)*dble(mtarg)/dble(imax)

c     Average all times
      teval=teval/ntimes
      tinit=tinit/ntimes
      tfast=tfast/ntimes
      tslow=tslow/ntimes


ccc      write(6,*) ' direct    fast  '
      error=0
      do 20 i=1,imax
ccc      write(6,*) exact(i),transform(i)
      error=max(abs(exact(i)-transform(i)),error)
20    continue

      sumstr=0
      do 30 j=1,nsorc
      sumstr=sumstr+abs(strength(j))
30    continue

c     Compute error bound and compare
      z=q-1.0d0/q
      th=z/(2.0d0-z)
      errbd=2.0d0*th**(nterms+1)
ccc      errbd=max(errbd,toler)

cc      write(6,*) ' nsorc,mtarg,nterms,q,toler '
cc      write(13,*) ' nsorc,mtarg,nterms,q,toler '
cc      write(6,*) nsorc,mtarg,nterms,q,toler
cc      write(13,*) nsorc,mtarg,nterms,q,toler
cc
cc      write(6,*) ' fast evaluation of ',nsorc,mtarg,' took ',tfast
cc      write(6,*) ' direct evaluation of ',nsorc,imax,' took ',t4-t3
cc      write(6,*) ' direct evaluation of ',
cc     >                    nsorc,mtarg,' would take ',tslow
cc
cc      write(6,*) ' error = ',error
cc      write(6,*) ' error/sumstr of strengths = ',error/sumstr
cc
cc      write(13,*) ' fast evaluation of ',nsorc,mtarg,' took ',tfast
cc      write(13,*) ' direct evaluation of ',nsorc,imax,' took ',t4-t3
cc      write(13,*) ' direct evaluation of ',
cc     >                    nsorc,mtarg,' would take ',tslow
cc
cc      write(13,*) ' error = ',error
cc      write(13,*) ' error/sumstr of strengths = ',error/sumstr
cc
      end

