      program laplace
c     Computes fast Laplace transform using symmetric series.

      implicit real*8(a-h,o-z)
      parameter (jsmax=40000,kbmax=400,icfmax=400000)
      real*8 source(jsmax),strength(jsmax),
     >       target(jsmax),transform(jsmax),
     >       coeffs(icfmax),coefft(icfmax),exact(100)
      integer itbox(jsmax),ions(-kbmax:kbmax),iont(-kbmax:kbmax)

c     Generate random sources and targets in specified range
      
1     continue
      smin=0.10d0
      smax=10.00d0
      tmin=0.10d0
      tmax=10.00d0
      write(6,*) ' enter smin,smax,tmin,tmax'
      write(13,*) ' enter smin,smax,tmin,tmax'
      read(5,*) smin,smax,tmin,tmax
      write(6,*) smin,smax,tmin,tmax
      write(13,*) smin,smax,tmin,tmax

      write(6,*) ' enter nsorc,mtarg,nterms,q,toler'
      write(13,*) ' enter nsorc,mtarg,nterms,q,toler'
      read(5,*) nsorc,mtarg,nterms,q,toler
      if(nsorc.le.0)goto 1000
      write(6,*) nsorc,mtarg,nterms,q,toler
      write(13,*) nsorc,mtarg,nterms,q,toler

      write(6,*) ' Generating points '
c     Generate targets
      idum=0
      call mkpoints(tmin,tmax,mtarg,target,strength,idum)
c     Generate sources and strengths
      call mkpoints(smin,smax,nsorc,source,strength,idum)
      write(6,*) ' smin,smax = ',smin,smax
      write(13,*) ' smin,smax = ',smin,smax
      write(6,*) ' tmin,tmax = ',tmin,tmax
      write(13,*) ' tmin,tmax = ',tmin,tmax

ccc      write(6,*) ' j          source(j)       strength(j)  '
ccc      write(13,*) ' j          source(j)       strength(j)  '
ccc      do 3 j=1,nsorc
ccc      write(6,*) j,source(j),strength(j)
ccc      write(13,*)j,source(j),strength(j)
ccc3     continue

ccc      write(6,*) 
ccc      write(13,*)
ccc      write(6,*) ' j          target(j)       '
ccc      write(13,*) ' j          target(j)       '
ccc      do 4 j=1,mtarg
ccc      write(6,*) j,target(j)
ccc      write(13,*) j,target(j)
ccc4     continue

c     Begin timing fast algorithm
      
      t1=second()

c     Form coefficients
      
ccc      write(6,*) ' Forming coefficients'
      call formcf(smin,smax,nsorc,nterms,q,kbsmin,kbsmax,
     >            source,strength,coeffs,ions,
     >               alogq,qsq,factor,kcut)

ccc      nsbox=kbsmax-kbsmin
ccc      write(6,*) ' number of source boxes is ',
ccc     >            nsbox,' = ',kbsmax ,' - ',kbsmin 
ccc      write(13,*) ' number of source boxes is ',
ccc     >            nsbox,' = ',kbsmax ,' - ',kbsmin 

c      do 10 n=1,100
c      write(6,*) coeffs(n)
c10    continue

c     Assign targets to boxes
      
ccc      write(6,*) ' Assigning targets to boxes'
      call assign(q,kctmin,kctmax,tmin,tmax,mtarg,
     >            target,itbox,iont,
     >               alogq,qsq,factor,kcut)

ccc      mtbox=kctmax-kctmin
ccc      write(6,*) ' number of target boxes is ',
ccc     >            mtbox,' = ',kctmax ,' - ',kctmin 
ccc      write(13,*) ' number of target boxes is ',
ccc     >            mtbox,' = ',kctmax ,' - ',kctmin 

ccc      ntpsq=(nterms+1)**2
ccc      if(nsbox*ntpsq .gt. icfmax .or.
ccc     >   mtbox*ntpsq .gt. icfmax .or.
ccc     >   nsbox .gt. kbmax .or.
ccc     >   mtbox .gt. kbmax)then
ccc      write(6,*) ' error in laplace transform : arrays too small '
ccc      stop
ccc      end if

c     Transform coeffs to coefft and accumulate

ccc      write(6,*) ' Transforming coefficients '
      call transt(kbsmin,kbsmax,kctmin,kctmax,nterms,q,toler,
     >               coeffs,coefft,ions,iont,
     >               alogq,qsq,factor,kcut)

c      do 11 n=1,40
ccc      write(6,*) coefft(n)
c11    continue

      t2=second()
      tinit=t2-t1
      t1=second()

c     Evaluate expansion at each target

ccc      write(6,*) ' Evaluating expansions'
      call eval(mtarg,nterms,kctmin,kctmax,q,
     >                target,coefft,iont,transform,itbox,
     >               alogq,qsq,factor,kcut)

c     Finish timing fast algorithm

      t2=second()
      teval=t2-t1
      tfast=tinit+teval

      nactb=0
      do 123 kbox=kctmin,kctmax
      nactb=nactb+iont(kbox)
123   continue
      write(6,*) ' number of active target boxes is ',nactb
      write(13,*) ' number of active target boxes is ',nactb

      nacsb=0
      do 125 kbox=kbsmin,kbsmax
      nacsb=nacsb+ions(kbox)
125   continue
      write(6,*) ' number of nonempty source boxes is ',nacsb
      write(13,*) ' number of nonempty source boxes is ',nacsb

c     Check that array bashing doesn't happen
      imax=min(100,mtarg)

c     Evaluate directly and compare to fast transform

      write(6,*) ' Direct evaluation '
      t3=second()
      call direct(nsorc,imax,source,strength,target,exact)
      t4=second()
      tslow=(t4-t3)*dble(mtarg)/dble(imax)

ccc      write(6,*) ' direct    fast  '
      error=0.0d0
      do 20 i=1,imax
ccc      write(6,*) exact(i),transform(i)
      error=max(dabs(exact(i)-transform(i)),error)
20    continue

      sum=0.0d0
      do 30 j=1,nsorc
      sum=sum+dabs(strength(j))
30    continue

c     Compare error bound with error
      z=q-1.0d0/q
      th=z/(2.0d0-z)
      errbd=2.0d0*th**(nterms+1)
      errbd=max(errbd,toler)

      write(6,*) ' nsorc,mtarg,nterms,q,toler '
      write(13,*) ' nsorc,mtarg,nterms,q,toler '
      write(6,*) nsorc,mtarg,nterms,q,toler
      write(13,*) nsorc,mtarg,nterms,q,toler

      write(6,*) ' initialization of ',nsorc,mtarg,' took ',tinit
      write(6,*) ' fast evaluation of ',nsorc,mtarg,' took ',tfast
      write(6,*) ' direct evaluation of ',nsorc,imax,' took ',t4-t3
      write(6,*) ' direct evaluation of ',
     >                    nsorc,mtarg,' would take ',tslow

      write(6,*) ' error = ',error
      write(6,*) ' error/sum of strengths = ',error/sum
      write(6,*) ' error bound = ',errbd

      write(6,*) '   '

      write(13,*) ' initialization of ',nsorc,mtarg,' took ',tinit
      write(13,*) ' fast evaluation of ',nsorc,mtarg,' took ',tfast
      write(13,*) ' direct evaluation of ',nsorc,imax,' took ',t4-t3
      write(13,*) ' direct evaluation of ',
     >                    nsorc,mtarg,' would take ',tslow

      write(13,*) ' error = ',error
      write(13,*) ' error/sum of strengths = ',error/sum
      write(13,*) ' error bound = ',errbd

      write(13,*) '   '

      goto 1
1000  continue
      end

****************************************************************************
      subroutine mkpoints(tmin,tmax,mtarg,target,strength,idum)
****************************************************************************
      implicit real*8(a-h,o-z)
      parameter (jsmax=40000)
      real*8 strength(jsmax),target(jsmax)
c     Generates random points between 0 and 1

      tmin1=1.0d+18
      tmax1=0.0d0
      if(idum.eq.0)then
      idum=-1
      x=ran1(idum)
      end if
      sqm=sqrt(dble(mtarg))
      do 10 j=1,mtarg
      ss=ran1(idum)
      target(j)=tmin+(tmax-tmin)*ss
ccc      target(j)=5.0d0*ss
      ss=ran1(idum)
      strength(j)=ss/sqm
      tmin1=min(tmin1,target(j))
      tmax1=max(tmax1,target(j))
10    continue
      tmin=tmin1
      tmax=tmax1

      return
      end
